
module Compute (Etat:Tools.Base)(Alph:Tools.Base) = struct

  type etat = Etat.t

  type lettre = Alph.t

  type automate = (etat,lettre) Automate.automate

  module Rel = Relation.Make(Etat)

  let (|>) = Rel.inser

  let (==) = Rel.equal

  let (@@) = Rel.concatenation

  let (++) = Rel.union

  module AMap = struct 

    module M = Map.Make(Alph)

    type key = lettre

    type t = Rel.t M.t

    let add = M.add

    let empty = M.empty

    let find a m =
      try M.find a m
      with Not_found -> Rel.empty

    let add_elt (a : lettre) (elt : etat * etat) (m : t) : t=
      M.add a (elt |> (find a m)) m

  end

  let base_rel (et,init,fn,alph,tr) =
    let m =
      List.fold_left (fun acc (q,a,q') -> AMap.add_elt a (q,q') acc) AMap.empty tr 
    in
    (fun a -> AMap.find a m)
    
  (*let g1 
      (base_rel : lettre -> Rel.t) (etats : etat list) 
      (alph : lettre list) (prime : lettre -> lettre) 
      : lettre -> Rel.t =
    let m =
      List.fold_left 
	(fun acc a -> 
	  AMap.add a (Rel.etoile etats ((base_rel (prime a))@@(base_rel a))) acc) 
	AMap.empty
	alph 
    in
    (fun a -> AMap.find a m)*)

  module MMap = struct 
    module M = Map.Make(Tools.ListB(Alph))
    type t = Rel.t M.t
    let empty = M.empty
    let add = M.add
    let find = M.find
    let choose = M.choose
    let filter = M.filter
  end

  type state = MMap.t

  let get p m =
    MMap.choose (MMap.filter p m)

  let g2 
      (base_rel : lettre -> Rel.t) 
      (etats : etat list)
      (prime : lettre -> lettre) 
      (s : state) 
      (w : lettre list)
      (a : lettre) 
      : state * lettre list * Rel.t =
    try (s, a::w, MMap.find (a::w) s)
    with Not_found -> 
      let rw = MMap.find w s
      and ra = base_rel a
      and ra' = base_rel (prime a) in
      let raw =	Rel.etoile etats (ra' @@ rw @@ ra) in
      try (s,fst (get (fun _ r -> r == raw) s),raw)
      with Not_found -> (MMap.add (a::w) raw s,a::w,raw)

  let trans_det
      (base_rel : lettre -> Rel.t) 
      (g2 : state -> lettre list -> lettre -> state * lettre list * Rel.t) 
      (s : state) 
      (q,w : etat list * lettre list) 
      (a : lettre) 
      : state * (etat list * lettre list) =
    match q with
      | [] -> (s,([],[]))
      | q -> 
	let s',w',r2 = g2 s w a in
	let r1 = base_rel a in
	let r1' = Rel.restr q r1 in
	match 
	  (List.fold_left 
	     (fun acc (a,q) -> Tools.inser Etat.compare q acc) 
	     [] 
	     (Rel.to_list (r1'@@r2))) with
	  | [] -> (s',([],[]))
	  | q' -> (s',(q',w'))
	(*let s',w',r2 = g2 s w a in
	let r1 = base_rel a in
	let r1' = Rel.restr q r1 in
	(s',(List.map snd (Rel.to_list (r1'@@r2)),w'))*)
  
  let auto_f_det (et,init,fn,alph,tr : automate) (prime : lettre -> lettre) =
    let alphabet = Tools.conc Alph.compare [List.map prime alph;alph] in
    let base_rel = base_rel (et,init,fn,alph,tr) in
    let g2 = g2 base_rel et prime in
    let fn = List.sort Etat.compare fn in
    let initial = (init,[]) 
    and final (q,_) = Tools.inter Etat.compare (List.sort Etat.compare q) fn 
    and trans = trans_det base_rel g2 in
    Automate.build_acc 
      (module Tools.PairB(Tools.ListB(Etat)) (Tools.ListB(Alph)))
      initial final alphabet trans (MMap.add [] (Rel.id et) MMap.empty)

  let trans
      (base_rel : lettre -> Rel.t) 
      (g2 : state -> lettre list -> lettre -> state * lettre list * Rel.t) 
      (s : state) 
      (q,w : etat * lettre list) 
      (a : lettre) 
      : state * (etat * lettre list) list =
    let s',w',r2 = g2 s w a in
    let r1 = base_rel a in
    let r1' = Rel.restr [q] r1 in
    (s',List.map (fun (_,q') -> (q',w')) (Rel.to_list (r1'@@r2)))
  
  let auto_f (et,init,fn,alph,tr : automate) (prime : lettre -> lettre) =
    let alphabet = Tools.conc Alph.compare [List.map prime alph;alph] in
    let base_rel = base_rel (et,init,fn,alph,tr) in
    let g2 = g2 base_rel et prime in
    let fn = List.sort Etat.compare fn in
    let initial = List.map (fun i -> i,[]) init 
    and final (q,_) = Tools.mem Etat.compare q fn 
    and trans = trans base_rel g2 in
    Automate.build2_acc 
      (module Tools.PairB(Etat) (Tools.ListB(Alph)))
      initial final alphabet trans (MMap.add [] (Rel.id et) MMap.empty)

end
(*

let get_g 
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat)
    (module Alph : Base with type t = lettre)
    (a : (etat,lettre) automate)
    (prime : lettre -> lettre)
    : lettre list -> (etat*etat) list
  =
 (* let (et,init,fn,alph,tr as a') = min2(module Etat)(module Alph) a in*)
  let module C = Compute (Etat) (Alph) in
  let trad,(elts,_ as m) = C.monsyn3 a in
  let g = C.func_g m prime in
  let find = List.fold_left (fun acc a -> g acc a) [] in
  let (|>) = inser (comp_to_pair Etat.compare Etat.compare) in
  let get_tab w =
    assoc (comp_to_list Alph.compare) w elts
  in
  let extract_pairs g =
    List.fold_left
      (fun acc w ->
	let t = get_tab w in
	fold_lefti 
	  (fun i acc -> function None -> acc | Some q->(trad i,q)|>acc)
	  acc
	  t)
      []
      g
  in
  (fun w -> extract_pairs (find w))

let auto_f 
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat)
    (module Alph : Base with type t = lettre)
    (_,init,fn,alph,tr as a : (etat,lettre) automate)
    (prime : lettre -> lettre)
    : (etat * lettre list,lettre) automate 
  =
  let q0 =
    match init with 
      | [q] -> (q,[]) 
      | _ -> failwith "auto_f : automate non déterministe" 
  in
  let g = get_g (module Etat)(module Alph) a prime in
  let ftr = makefuncd (module Etat)(module Alph) tr in
  let trans (q,w) a =
    match ftr q a with
      | None -> []
      | Some q' ->
	let w' = (w@[a]) in
	List.map (fun q -> (q,w')) (assoc2 Etat.compare q' (g w'))
  in
  let fn' = List.sort Etat.compare fn in
  let final (q,_) = mem Etat.compare q fn' in
  build2(module PairB(Etat)(ListB(Alph))) q0 final alph trans 

let read_from 
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat)
    (trans : etat -> lettre -> etat list) =
  let f = 
    (fun ql a -> bind Etat.compare (fun q -> trans q a) ql)
  in
  (fun q w ->
    List.fold_left f [q] w)

(*let read_star
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat) 
    (read : etat -> lettre list -> etat list)
    (etats : etat list)
    (w : lettre list) =
  let (<<) = comp_to_pair Etat.compare Etat.compare in
  let (|>) = inser (<<) in
  let etats' = (List.sort Etat.compare etats) in
  let refl = List.map (fun q -> (q,q)) etats' in
  let step = 
    List.fold_left 
      (fun acc q -> 
	List.fold_left 
	  (fun acc q' -> 
	    (q,q') |> acc) 
	  acc 
	  (read q w)) 
      [] 
      (List.rev etats') 
  in
  let rec augm b dejavu next (q,q') = function
    | [] -> (b,dejavu,next)
    | (q1,q2)::todo when q = q2 & q1 <> q' ->
      begin
	match (inserb (<<) (q1,q') dejavu) with
	  | None -> augm b dejavu next (q,q') todo
	  | Some dejavu' -> augm true dejavu' ((q1,q')|> next) (q,q') todo
      end
    | (q1,q2)::todo when q'= q1 & q2 <>q ->
      begin
	match (inserb (<<) (q,q2) dejavu) with
	  | None -> augm b dejavu next (q,q') todo
	  | Some dejavu' -> augm true dejavu' ((q,q2)|> next) (q,q') todo
      end
    | _::todo -> augm b dejavu next (q,q') todo
  in
  let rec aux dejavu = function
    | [] -> dejavu
    | (q,q')::todo -> 
      let _,dejavu',next = augm false ((q,q')|>dejavu) todo (q,q') todo in
      aux dejavu' next
  in
  let cltr = aux step step in
  conc (<<) [cltr;refl]*)

let concat
    (type elt)
    (module Elts : Base with type t = elt)
    (ens1 : (elt * elt) list)
    (ens2 : (elt * elt) list)
    (context : (elt * elt) list)
    : (elt * elt) list * (elt * elt) list=
  let comp = comp_to_pair Elts.compare Elts.compare in
  let ens1' = List.sort (comp_to_pair2 Elts.compare Elts.compare) ens1 in
  let ens2' = List.sort (comp_to_pair Elts.compare Elts.compare) ens2 in
  let tranche1 =
    let rec aux acc x = function
      | [] -> [x,acc]
      | (y,x')::q when x'=x -> aux (y::acc) x q
      | (y,x')::q -> (x,acc)::(aux [y] x' q)
    in
    match ens1' with
      | [] -> []
      | (y,x)::q -> aux [y] x q
  and tranche2 =
    let rec aux acc x = function
      | [] -> [x,acc]
      | (x',y)::q when x'=x -> aux (y::acc) x q
      | (x',y)::q -> (x,acc)::(aux [y] x' q)
    in
    match ens2' with
      | [] -> []
      | (x,y)::q -> aux [y] x q
  in
  let mix e1 e2 acc =
    List.fold_left
      (fun acc x ->
	List.fold_left
	  (fun (acc1,acc2) y -> 
	    match inserb comp (x,y) acc1 with
	      | None -> (acc1,acc2)
	      | Some acc1' -> (acc1',(x,y)::acc2))
	  acc
	  e2)
      acc
      e1
  in
  let rec aux = function
    | (x1,t1)::q1,(x2,t2)::q2 ->
      begin
	match Elts.compare x1 x2 with
	  | 0 -> mix t1 t2 (aux (q1,q2))
	  | k when k>0 -> aux ((x1,t1)::q1,q2)
	  | _ -> aux (q1,(x2,t2)::q2)
      end
    | _ -> (context,[])
  in
  aux (tranche1,tranche2)

let concatenation
    (type elt)
    (module Elts : Base with type t = elt)
    (ens1 : (elt * elt) list)
    (ens2 : (elt * elt) list)
    : (elt * elt) list=
  let ens1' = List.sort (comp_to_pair2 Elts.compare Elts.compare) ens1 in
  let ens2' = List.sort (comp_to_pair Elts.compare Elts.compare) ens2 in
  let tranche1 =
    let rec aux acc x = function
      | [] -> [x,acc]
      | (y,x')::q when x'=x -> aux (y::acc) x q
      | (y,x')::q -> (x,acc)::(aux [y] x' q)
    in
    match ens1' with
      | [] -> []
      | (y,x)::q -> aux [y] x q
  and tranche2 =
    let rec aux acc x = function
      | [] -> [x,acc]
      | (x',y)::q when x'=x -> aux (y::acc) x q
      | (x',y)::q -> (x,acc)::(aux [y] x' q)
    in
    match ens2' with
      | [] -> []
      | (x,y)::q -> aux [y] x q
  in
  let mix e1 e2 acc =
    List.fold_left
      (fun acc x ->
	List.fold_left
	  (fun acc y -> (x,y)::acc)
	  acc
	  e2)
      acc
      e1
  in
  let rec aux = function
    | (x1,t1)::q1,(x2,t2)::q2 ->
      begin
	match Elts.compare x1 x2 with
	  | 0 -> mix t1 t2 (aux (q1,q2))
	  | k when k>0 -> aux ((x1,t1)::q1,q2)
	  | _ -> aux (q1,(x2,t2)::q2)
      end
    | _ -> []
  in
  aux (tranche1,tranche2)

let union 
    (type elt)
    (module Elts : Base with type t = elt)
    (ens1 : (elt * elt) list)
    (ens2 : (elt * elt) list)
    : (elt * elt) list =
  conc (comp_to_pair Elts.compare Elts.compare) [ens1;ens2]

let etoile
    (type elt)
    (module Elts : Base with type t = elt)
    (ens : (elt * elt) list)
    (elts : elt list)
    : (elt * elt) list =
  let rec aux current = function
    | [] -> current
    | todo ->
      let (nouv,next) = concat (module Elts) todo ens current in
      aux nouv next
  in
  let init = 
    conc 
      (comp_to_pair Elts.compare Elts.compare) 
      [List.map (fun x -> x,x) elts;ens]
  in
  aux init ens

let g1
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat)
    (module Alph : Base with type t = lettre)
    (read : etat -> lettre list -> etat list)
    (etats : etat list)
    (alph : lettre list)
    (prime : lettre -> lettre)
    : lettre -> (etat*etat) list =
  let module Map = Map.Make(Alph) in
  let map = 
    List.fold_left
      (fun m a ->
	Map.add a 
	  (etoile 
	     (module Etat) 
	     (List.fold_left 
		(fun acc q -> 
		  (List.map (fun q' -> q,q') (read q [prime a;a]))@acc)
		[]
		etats) 
	     etats) 
	  m)
      Map.empty
      alph
  in
  (fun a -> Map.find a map)

let g2
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat)
    (module Alph : Base with type t = lettre)
    (read : lettre -> (etat*etat) list)
    (etats : etat list)
    (alph : lettre list)
    (prime : lettre -> lettre)
    (g1 : lettre -> (etat*etat) list)
    : lettre list -> (etat*etat) list = 
  let (@@) = concatenation(module Etat)
  and (++) = union(module Etat)
  and etoile = etoile(module Etat) in
  let rec aux = function
    | [] -> []
    | a::w -> 
      let r = aux w 
      and g = g1 a in
      ((read (prime a))@@r@@(etoile (g@@r) etats)@(read a))++g
  in
  (fun w -> aux (List.rev w))*)
      

      
  

(*let etoile
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat) 
    (read : etat -> lettre list -> etat list)
    (etats : etat list)
    (w : lettre list) = 
  let module UF = MakeUF(Etat)(Etat) in
  let etats' =
    List.sort (fun q q' -> Etat.compare q' q) etats
  in
  let step state q = 
    List.fold_left 
      (fun acc q' -> fst (UF.union (A q) (B q') acc ()))
      state
      (read q w)
  in
  let (=>) =
    let state = List.fold_left step UF.initial etats in
    (fun q q' -> UF.equivalent (A q) (B q') state)
  in
  List.fold_left
    (fun acc q ->
      List.fold_left
	(fun acc q' ->
	  if q=>q'
	  then (q,q')::acc
	  else acc)
	acc
	etats')
    []
    etats'*)


(*let make_g1 
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat)
    (module Alph : Base with type t = lettre)
    (_,_,_,_,tr : (etat,lettre) automate) =
    let f = makefuncnd (module Etat)(module Alph) tr in*)
