open Regexp
open Automate

module Str = struct
  type t = string
  let compare = compare
end

module Int = struct
  type t = int
  let compare = compare
end

module Lang = Tools.ListB(Tools.ListB(Str))

module PLang = Tools.PairB(Lang)(Lang)

module C = Monoid.Compute(Int)(Str)

module Etat = Tools.OptionB(Tools.PairB(Str)(Int))

module DEtat = Tools.PairB(Tools.ListB(Etat))(Tools.ListB(Str))

module NDEtat = Tools.PairB(Etat)(Tools.ListB(Str))

module DEtatb = Tools.PairB(Tools.ListB(Int))(Tools.ListB(Str))

module NDEtatb = Tools.PairB(Int)(Tools.ListB(Str))

module CR = Construction.Compute(Etat)(Str)

module CRb = Construction.Compute(Int)(Str)

(** Arguments *)


let file,pr,drinit,drf,gauto,timed,run1,run2,run3,psize,nbexpo = 
  let file = ref ""  
  and pr = ref false 
  and drinit = ref false
  and drf = ref false 
  and gauto = ref false 
  and timed = ref false 
  and run1 = ref true
  and run2 = ref true 
  and run3 = ref true 
  and psize = ref false 
  and nb = ref 0 in
  Arg.parse [
    ("-expo",Arg.Set_int nb,"Special mode, to test exponential automaton. Specify the size of the automaton. Cancels other modes!");
    ("-print",Arg.Set pr,"Print the monoids.");
    ("-not1",Arg.Clear run1,"Do not try the first method.");
    ("-not2",Arg.Clear run2,"Do not try the second method.");
    ("-not3",Arg.Clear run3,"Do not try the third method.");
    ("-draw-all",
     Arg.Unit 
       (fun () -> 
	 drinit := true;drf := true)
      ,"Draw all of the automata.");
    ("-draw-i",Arg.Set drinit,"Draw the initial automata.");
    ("-draw-f",Arg.Set drf,"Draw the 'f' automata.");
    ("-g-auto",Arg.Set gauto,"Construct and draw the 'g' automata.");
    ("-chrono",Arg.Set timed,"Print time information.");
    ("-size",Arg.Set psize,"Print the size of the 'f' automata.");
    ("-v",Arg.Unit 
      (fun () -> timed:=true;psize:=true;drinit := true;drf := true),
     "Shorthand for -chrono -size");
    ("-1",Arg.Unit 
      (fun () -> run2:=false;run3:=false),
     "Only run the first method");
    ("-2",Arg.Unit 
      (fun () -> run1:=false;run3:=false),
     "Only run the second method");
    ("-3",Arg.Unit 
      (fun () -> run2:=false;run1:=false),
     "Only run the third method")
    ] 
      (fun s -> file:=s) 
      "Specify the name of the file to process.";
  (!file,!pr,!drinit,!drf,!gauto,!timed,!run1,!run2,!run3,!psize,!nb)


(**************************************)


let print_lettre a = a

let print_mot = Monoid.mot_to_string print_lettre

let print_intlst (ql,h) = (Monoid.list_to_string string_of_int ql)^","^(print_mot h)

let print_int (q,h) = (string_of_int q)^","^(print_mot h)

let print_etat = Monoid.pos_to_string print_lettre

let print_lang = Monoid.lang_to_string print_lettre

let printpairlang (f,g) =  (print_lang f)^","^(print_lang g)

let printdetat (ql,w) = 
  "["^(String.concat ";" (List.map print_etat ql))^"],"^(print_mot w)

let printndetat (q,w) = 
  (print_etat q)^","^(print_mot w)

let printm =
  if pr 
  then (Some (print_lettre,string_of_int)) 
  else None

let prime a = 
  if (a.[String.length a - 1] = '\'') 
  then (String.sub a 0 (String.length a - 1))
  else a^"'"

let full_alph alph =
  let inser = Tools.inser Str.compare in
  List.fold_left (fun l a -> inser a (inser (prime a) l)) [] alph

let get_alph e =
  let inser = Tools.inser Str.compare in
  snd (fold (fun l a -> (a,inser a l)) [] e)

(*let lequiv = equiv (module PLang) (module PLang) (module Str)*)

let lhkequiv = hkequiv (module PLang) (module PLang) (module Str)

(*let lequiv2 = equiv (module DEtat) (module DEtat) (module Str)*)

let lhkequiv2 = hkequiv (module DEtat) (module DEtat) (module Str)

(*let lequiv3 = equiv (module DEtat) (module DEtat) (module Str)*)

let lhkequiv3 = hkequiv (module NDEtat) (module NDEtat) (module Str)

let lhkequiv2b = hkequiv (module DEtatb) (module DEtatb) (module Str)

let lhkequiv3b = hkequiv (module NDEtatb) (module NDEtatb) (module Str)

let draw b name print_etat print_lettre auto =
  if b
  then 
    begin
      print_to_file name print_etat print_lettre auto;
      let _ = Sys.command ("dot -Tpng "^name^".dot -o "^name^".png") in
      let _ = Sys.command ("mv "^name^".png Dessin") in
      let _ = Sys.command ("rm "^name^".dot") in
      ()
    end
  else ()

let print_size indice op nb vers (e,_,_,_,t) =
  if psize
  then 
    (Printf.printf "%d%s--%d-%d : Size = %d états, %d transitions." 
       indice op vers nb (List.length e) (List.length t);
     print_newline())
  else ()

let make_auto op indice nb e =
  Printf.printf "%d%s : Construction de l'automate %d." indice op nb;
  let name = (file^"-"^(string_of_int indice)^op^"-"^(string_of_int nb)) in
  let e' = rev_to_reg prime e in
  let alph = full_alph (get_alph e') in
  let auto = auto_r Str.compare alph e' in
  draw drinit name print_etat print_lettre auto;
  print_newline();
  (alph,auto)

let test_eq1 indice op (res,(alph1,auto1),(alph2,auto2)) =
  let pref s = (file^"-"^(string_of_int indice)^op^"-"^s^"-1") in
  let auto1m = simpl (min2 (module Int) (module Str) (simpl auto1)) 
  and auto2m = simpl (min2 (module Int) (module Str) (simpl auto2))
  in
  let p1,l1,m1 = C.monsyn2 printm auto1m
  and p2,l2,m2 = C.monsyn2 printm auto2m in
  if gauto
  then
    begin
      let g1 = C.auto_g m1 p1 alph1 prime 
      and g2 = C.auto_g m2 p2 alph2 prime 
      in
      draw true (pref "1g") print_lang print_lettre g1;
      draw true (pref "2g") print_lang print_lettre g2
    end;
  let f1 = C.auto_f_opt3 m1 p1 l1 alph1 prime 
  and f2 = C.auto_f_opt3 m2 p2 l2 alph2 prime in
  print_size indice op 1 1 f1;
  print_size indice op 2 1 f2;
  draw drf (pref "1f") printpairlang print_lettre f1;
  draw drf (pref "2f") printpairlang print_lettre f2;
  let(* b = lequiv f1 f2
  and*) b = lhkequiv f1 f2
  in
  ((match (res,b) with
    | Some true,None | Some false,Some _ -> 
      Printf.printf "%d%s--1 : OK." indice op
    | Some false,None -> 
      Printf.printf 
	"%d%s--1 : L'équation est fausse : les automates sont équivalents." 
	indice op
    | Some true,Some u -> 
      Printf.printf 
	"%d%s--1 : L'équation est fausse : considérons le mot %s." 
	indice op (print_mot u)
    | None,None -> 
      Printf.printf "%d%s--1 : Les deux expressions sont équivalentes." indice op
    | None,Some u -> 
      Printf.printf 
	"%d%s--1 : Les deux expressions ne sont pas équivalentes : considérons le mot %s."
	indice op (print_mot u));
   print_newline ())

let test_eq2 indice op (res,(alph1,auto1),(alph2,auto2)) =
  let pref s = (file^"-"^(string_of_int indice)^op^"-"^s^"-2") in
  let f1 = CR.auto_f_det auto1 prime 
  and f2 = CR.auto_f_det auto2 prime in
  print_size indice op 1 2 f1;
  print_size indice op 2 2 f2;
  draw drf (pref "1f") printdetat print_lettre f1;
  draw drf (pref "2f") printdetat print_lettre f2;
  let auto1m = simpl (min2 (module Int) (module Str) (simpl f1)) 
  and auto2m = simpl (min2 (module Int) (module Str) (simpl f2))
  in  
  draw drf (pref "1mf") string_of_int print_lettre auto1m;
  draw drf (pref "2mf") string_of_int print_lettre auto2m;
  let (*b = lequiv2 f1 f2
  and*) b = lhkequiv2 f1 f2
  in
  ((match (res,b) with
    | Some true,None | Some false,Some _ -> 
      Printf.printf "%d%s--2 : OK." indice op
    | Some false,None -> 
      Printf.printf 
	"%d%s--2 : L'équation est fausse : les automates sont équivalents." 
	indice op
    | Some true,Some u -> 
      Printf.printf 
	"%d%s--2 : L'équation est fausse : considérons le mot %s." 
	indice op (print_mot u)
    | None,None -> 
      Printf.printf "%d%s--2 : Les deux expressions sont équivalentes." indice op
    | None,Some u -> 
      Printf.printf 
	"%d%s--2 : Les deux expressions ne sont pas équivalentes : considérons le mot %s."
	indice op (print_mot u));
   print_newline ())

let test_eq3 indice op (res,(alph1,auto1),(alph2,auto2)) =
  let pref s = (file^"-"^(string_of_int indice)^op^"-"^s^"-3") in
  let f1 = CR.auto_f auto1 prime 
  and f2 = CR.auto_f auto2 prime in
  print_size indice op 1 3 f1;
  print_size indice op 2 3 f2;
  draw drf (pref "1f") printndetat print_lettre f1;
  draw drf (pref "2f") printndetat print_lettre f2;
  let (*b = lequiv3 f1 f2
  and*) b = lhkequiv3 f1 f2
  in
  ((match (res,b) with
    | Some true,None | Some false,Some _ -> 
      Printf.printf "%d%s--3 : OK." indice op
    | Some false,None -> 
      Printf.printf 
	"%d%s--3 : L'équation est fausse : les automates sont équivalents." 
	indice op
    | Some true,Some u -> 
      Printf.printf 
	"%d%s--3 : L'équation est fausse : considérons le mot %s." 
	indice op (print_mot u)
    | None,None -> 
      Printf.printf "%d%s--3 : Les deux expressions sont équivalentes." indice op
    | None,Some u -> 
      Printf.printf 
	"%d%s--3 : Les deux expressions ne sont pas équivalentes : considérons le mot %s."
	indice op (print_mot u));
   print_newline ())
  

let test_eq =
  if timed
  then 
    (fun i o (r,e1,e2) -> 
      let e = (r,Tools.chrono (make_auto o i 1) e1,Tools.chrono (make_auto o i 2) e2) 
      in
      if run1 then Tools.chrono (test_eq1 i o) e; 
      if run2 then Tools.chrono (test_eq2 i o) e; 
      if run3 then Tools.chrono (test_eq3 i o) e)
  else 
    (fun i o (r,e1,e2) -> 
      let e = (r,make_auto o i 1 e1,make_auto o i 2 e2) in
      if run1 then test_eq1 i o e;
      if run2 then test_eq2 i o e;
      if run3 then test_eq3 i o e)

let process indice (cl,e1,e2) =
  match cl with
    | `Eq -> test_eq indice "" (Some true,e1,e2)
    | `Neq -> test_eq indice "" (Some false,e1,e2)
    | `Leq -> test_eq indice "" (Some true,RUnion (e1,e2),e2)
    | `Geq -> test_eq indice "" (Some true,e1,RUnion (e1,e2))
    | `Dunno -> test_eq indice "" (None,e1,e2)
    | `Lt -> 
      (test_eq indice "-leq" (Some true,RUnion (e1,e2),e2);
       test_eq indice "-diff" (Some false,e1,e2))
    | `Gt -> 
      (test_eq indice "-geq" (Some true,e1,RUnion (e1,e2));
       test_eq indice "-diff" (Some false,e1,e2))
    | `Incomp ->
      (test_eq indice "-nleq" (Some false,RUnion (e1,e2),e2);
       test_eq indice "-ngeq" (Some false,e1,RUnion (e1,e2)))



(*let gen_n n =
  let rec aux acc1 acc2 = function
    | 0 ->  (0::acc1,(0,"a",0)::(0,"a'",0)::(0,"a'",n+2)::(n+2,"a'",1)::acc2)
    | k -> (aux (k::acc1) ((k,"a",k+1)::(k,"a'",k+1)::acc2) (k-1))
  in
  let (e,tr) = aux [n+1;n+2] [] n in
  (e,[0],[n+1],["a";"a'"],tr)*)

let gen_n n =
  let rec aux acc1 acc2 = function
    | 0 ->  (0::acc1,(0,"b",0)::(0,"a'",0)::(0,"a",0)::(0,"a",1)::acc2)
    | k -> (aux (k::acc1) ((k,"a",k+1)::acc2) (k-1))
  in
  let rec aux2 acc1 acc2 = function
    | k when k=n -> (aux (k::acc1) ((k,"a",k+1)::(k,"b",k+1)::acc2) (k-1))
    | k -> (aux2 (k::acc1) ((k,"a",k+1)::(k,"b",k+1)::acc2) (k-1))
  in
  let (e,tr) = aux2 [2*n] [] (2*n-1) in
  (List.rev e,[0],[2*n],["a";"a'";"b";"b'"],tr)

(*let gen_ab n =
  let rec aux acc1 acc2 = function
    | 0 ->  (0::acc1,(0,"a",0)::(0,"a'",0)::(0,"a'",n+2)::(n+2,"a'",1)::acc2)
    | k -> (aux (k::acc1) ((k,"a",k+1)::(k,"a'",k+1)::acc2) (k-1))
  in
  let (e,tr) = aux [n+1;n+2] [] n in
  (e,[0],[n+1],["a";"a'";"b";"b'"],tr)*)

let tauto =
  ([0;1],[0],[1],["a";"a'"],
   [(0,"a",1);
    (0,"a'",1);
    (1,"a",1);
    (1,"a'",1)])


let test_eq1b indice op (res,(alph1,auto1),(alph2,auto2)) =
  let pref s = (file^"-"^(string_of_int indice)^op^"-"^s^"-1") in
  let auto1m = simpl (min2 (module Int) (module Str) (simpl auto1)) 
  and auto2m = simpl (min2 (module Int) (module Str) (simpl auto2))
  in
  let p1,l1,m1 = C.monsyn2 printm auto1m
  and p2,l2,m2 = C.monsyn2 printm auto2m in
  if gauto
  then
    begin
      let g1 = C.auto_g m1 p1 alph1 prime 
      and g2 = C.auto_g m2 p2 alph2 prime 
      in
      draw true (pref "1g") print_lang print_lettre g1;
      draw true (pref "2g") print_lang print_lettre g2
    end;
  let f1 = C.auto_f_opt3 m1 p1 l1 alph1 prime 
  and f2 = C.auto_f_opt3 m2 p2 l2 alph2 prime in
  print_size indice op 1 1 f1;
  print_size indice op 2 1 f2;
  draw drf (pref "1f") printpairlang print_lettre f1;
  draw drf (pref "2f") printpairlang print_lettre f2;
  let(* b = lequiv f1 f2
  and*) b = lhkequiv f1 f2
  in
  ((match (res,b) with
    | Some true,None | Some false,Some _ -> 
      Printf.printf "%d%s--1 : OK." indice op
    | Some false,None -> 
      Printf.printf 
	"%d%s--1 : L'équation est fausse : les automates sont équivalents." 
	indice op
    | Some true,Some u -> 
      Printf.printf 
	"%d%s--1 : L'équation est fausse : considérons le mot %s." 
	indice op (print_mot u)
    | None,None -> 
      Printf.printf "%d%s--1 : Les deux expressions sont équivalentes." indice op
    | None,Some u -> 
      Printf.printf 
	"%d%s--1 : Les deux expressions ne sont pas équivalentes : considérons le mot %s."
	indice op (print_mot u));
   print_newline ())

let test_eq2b indice op (res,(alph1,auto1),(alph2,auto2)) =
  let pref s = (file^"-"^(string_of_int indice)^op^"-"^s^"-2") in
  let f1 = CRb.auto_f_det auto1 prime 
  and f2 = CRb.auto_f_det auto2 prime in
  print_size indice op 1 2 f1;
  print_size indice op 2 2 f2;
  draw drf (pref "1f") print_intlst print_lettre f1;
  draw drf (pref "2f") print_intlst print_lettre f2;
  let (*b = lequiv2 f1 f2
  and*) b = lhkequiv2b f1 f2
  in
  ((match (res,b) with
    | Some true,None | Some false,Some _ -> 
      Printf.printf "%d%s--2 : OK." indice op
    | Some false,None -> 
      Printf.printf 
	"%d%s--2 : L'équation est fausse : les automates sont équivalents." 
	indice op
    | Some true,Some u -> 
      Printf.printf 
	"%d%s--2 : L'équation est fausse : considérons le mot %s." 
	indice op (print_mot u)
    | None,None -> 
      Printf.printf "%d%s--2 : Les deux expressions sont équivalentes." indice op
    | None,Some u -> 
      Printf.printf 
	"%d%s--2 : Les deux expressions ne sont pas équivalentes : considérons le mot %s."
	indice op (print_mot u));
   print_newline ())

let test_eq3b indice op (res,(alph1,auto1),(alph2,auto2)) =
  let pref s = (file^"-"^(string_of_int indice)^op^"-"^s^"-3") in
  let f1 = CRb.auto_f auto1 prime 
  and f2 = CRb.auto_f auto2 prime in
  print_size indice op 1 3 f1;
  print_size indice op 2 3 f2;
  draw drf (pref "1f") print_int print_lettre f1;
  draw drf (pref "2f") print_int print_lettre f2;
  let auto1m = simpl (min2 (module Int) (module Str) (simpl f1)) 
  and auto2m = simpl (min2 (module Int) (module Str) (simpl f2))
  in  
  draw drf (pref "1mf") string_of_int print_lettre auto1m;
  draw drf (pref "2mf") string_of_int print_lettre auto2m;
  let (*b = lequiv3 f1 f2
  and*) b = lhkequiv3b f1 f2
  in
  ((match (res,b) with
    | Some true,None | Some false,Some _ -> 
      Printf.printf "%d%s--3 : OK." indice op
    | Some false,None -> 
      Printf.printf 
	"%d%s--3 : L'équation est fausse : les automates sont équivalents." 
	indice op
    | Some true,Some u -> 
      Printf.printf 
	"%d%s--3 : L'équation est fausse : considérons le mot %s." 
	indice op (print_mot u)
    | None,None -> 
      Printf.printf "%d%s--3 : Les deux expressions sont équivalentes." indice op
    | None,Some u -> 
      Printf.printf 
	"%d%s--3 : Les deux expressions ne sont pas équivalentes : considérons le mot %s."
	indice op (print_mot u));
   print_newline ())
  

let run file = 
  if nbexpo>0
  then
    if timed
    then  
      begin 
	let automate = gen_n nbexpo in
	draw drinit file  string_of_int print_lettre automate;
	if run1 then 
	  Tools.chrono 
	    (test_eq1b 0 "") 
	    (None,(["a";"a'"],automate),(["a";"a'"],tauto));      
	if run2 then  
	  Tools.chrono 
	    (test_eq2b 0 "") 
	    (None,(["a";"a'"],automate),(["a";"a'"],tauto));
	if run3 then
	  Tools.chrono 
	    (test_eq3b 0 "")
	    (None,(["a";"a'"],automate),(["a";"a'"],tauto))
      end
    else 
      begin 
	let automate = gen_n nbexpo in
	draw drinit file  string_of_int print_lettre automate;
	if run1 then 
	  test_eq1b 0 "" (None,(["a";"a'"],automate),(["a";"a'"],tauto));      
	if run2 then   
	  test_eq2b 0 "" (None,(["a";"a'"],automate),(["a";"a'"],tauto));
	if run3 then 
	  test_eq3b 0 "" (None,(["a";"a'"],automate),(["a";"a'"],tauto))
      end
  else
    begin
      let ch = open_in file in 
      let l = Lexing.from_channel ch in
      let eql = Expr_parser.eqs Expr_lexer.eqs_token l in
      close_in ch;
      let _ = List.fold_left (fun i eq -> process i eq; i+1) 1 eql in ()
    end

let _ = 
  if timed 
  then Tools.chrono run file
  else run file
