open Tools
open Automate

type ('a,'b) monoid =  ('a mot*'b array) list * ('a mot*'a mot) list


let mot_to_string f : 'a mot -> string = 
  let rec aux = function
    | [] -> ""
    | i::q -> (f i)^(aux q)
  in
  (function [] -> "ε" | l -> aux l)

let list_to_string g = function
  | [] -> "⊥"
  | l -> "["^(String.concat ";" (List.map g l))^"]" 

let table_to_string (g : 'b -> string) t  =
  Array.fold_right 
    (fun i s -> (g i)^" "^s) 
    t ""

let lang_to_string f l = String.concat "+" (List.map (mot_to_string f) l)

let pos_to_string f = function
  | None -> "⊥"
  | Some (a,i) -> (f a)^"_"^(string_of_int i)

let option_to_string f = function
  | None -> "⊥"
  | Some a -> (f a)

let printmon f g (m,e,p) =
  Printf.printf "\n ------------------------------- \n";
  Printf.printf "Table des éléments :\n";
  List.iter
    (fun (u,t) -> 
      Printf.printf "%s : %s\n" (mot_to_string f u) (table_to_string g t))
    m;
  Printf.printf "\nÉquivalences :\n";
  List.iter
    (fun (u,v) -> 
      Printf.printf "%s --> %s\n" (mot_to_string f u) (mot_to_string f v))
    e;
  Printf.printf "\nÉléments finaux :\n%s"
    (list_to_string (mot_to_string f) p);
  Printf.printf "\n ------------------------------- \n"

let augmsig c e tr (u,t) a =
  let t' = Array.make e None in
  for i = 0 to (e-1) do
    match t.(i) with
      | None -> ()
      | Some q -> (t'.(i)<- (tr q a))
  done;
  (a::u,t')

let montaille c comp e alph tr elts prec eq =
  let (<<) = comp in
  let rec aux t = function
    | [] -> A [t]
    | (u,_ as t')::q ->
      begin
	match (t<<t') with
	  | -1 -> A (t::t'::q)
	  | 1  -> 
	    begin
	      match (aux t q) with
		| A l -> A (t'::l)
		| B u -> B u
	    end
	  | _  -> B u
      end
  in
  let rec aux2 pasfini elts' prec' equiv = function
    | _,[] -> (pasfini,elts',prec',equiv)
    | [],t::q -> aux2 pasfini elts' prec' equiv (alph,q)
    | k::al,(u,_ as t)::q -> 
      begin
	let t'=(augmsig c e tr t k) in
	match (aux t' elts') with
	  | A l -> aux2 true l (t'::prec') equiv (al,t::q)
	  | B v -> aux2 pasfini elts' prec' ((k::u,v)::equiv) (al,t::q)
      end
  in
  aux2 false elts [] eq (alph,prec)

let rec (-|) a = function
  | [] -> A []
  | ([b],m)::_ when b=a -> (B m)
  | (b::r,m)::q when b=a -> 
    begin
      match (a -| q) with
	| A l -> A ((r,m)::l)
	| B m -> B m
    end
  | _::q ->  (a -| q)

let beta rules (m : 'a mot) : 'a mot= 
  let rec aux r acc = function
    | [] -> List.rev acc
    | a::m -> 
      begin
	match (a -| r) with
	  | A [] -> acc@@(a::(aux rules [] m))
	  | A r' -> (aux r' (a::acc) m)
	  | B m' -> m'@(aux rules [] m)
      end
  in
  aux rules [] m

module Compute (Etat : Base) (Alph : Base) = struct
  type etat = Etat.t
  type lettre = Alph.t
  type mot = lettre list

  let comp1 = Alph.compare
  let comp2 = Etat.compare
  let compmots = comp_to_list comp1
  let (|>) = inser compmots 
  let is_in = mem compmots

  module EMap = Map.Make(Etat)
  module Mot = ListB(Alph)

  let monsyn 
      etats_acc
      print
      ((e,i,fn,a,tr) : (etat,lettre) automate) 
      : mot list * mot list * (lettre,etat option) monoid =
    let q0 = 
      match i with 
	| [q] -> q 
	| _ -> failwith "monsyn : automate non déterministe" 
    in
    let n = List.length e in
    let tr' = 
      let f = makefuncnd(module Etat)(module Alph) tr in
      (fun q a ->
	match f q a with
	  | [] -> None
	  | [q'] -> Some q'
	  | _ -> failwith "caca")
    in
    let comp (_,t1) (_,t2) = comp_to_array (comp_to_option comp2) n t1 t2 in
    let rec step elts prec equiv =
      match (montaille comp2 comp n a tr' elts prec equiv) with
	| (false,_,_,eq) -> (elts,eq)
	| (_,a,p,e) -> step a p e
    in
    let init,trad = 
      let t' = Array.make n None in
      let m,_ =
	List.fold_left
	  (fun (m,k) q -> t'.(k)<-(Some q); (EMap.add q k m,k+1))
	  (EMap.empty,0)
	  e
      in
      (([],t'),fun q -> EMap.find q m)
    in
    let (elts,eq) = step [init] [init] []
    in
    let elts',eq' =
      (List.sort (fun (u,_) (v,_) -> compmots u v)
	 (List.map (fun (m,s) -> (List.rev m,s)) elts),
       List.map (fun (m,m') -> (List.rev m,List.rev m')) eq)
    in
    let p,l =
      let aux (acc1,acc2) (x,t) = 
	match t.(trad q0) with
	  | None -> (acc1,acc2)
	  | Some q' ->
	    if mem comp2 q' fn
	    then (inser compmots x acc1,inser compmots x acc2)
	    else 
	      if mem comp2 q' etats_acc
	      then (acc1,inser compmots x acc2)
	      else (acc1,acc2)
      in
      List.fold_left aux ([],[]) elts'
    in
    ((match print with
      | Some (f,g) -> printmon f (option_to_string g) (elts',eq',p)
      | None -> ());
     (p,l,(elts',eq')))

  let monsyn2 
      print
      ((e,i,fn,a,tr) : (etat,lettre) automate) 
      : mot list * mot list * (lettre,etat option) monoid =
    let q0 = 
      match i with 
	| [q] -> q 
	| _ -> failwith "monsyn : automate non déterministe" 
    in
    let n = List.length e in
    let tr' = 
      let f = makefuncnd(module Etat)(module Alph) tr in
      (fun q a ->
	match f q a with
	  | [] -> None
	  | [q'] -> Some q'
	  | _ -> failwith "caca")
    in
    let comp (_,t1) (_,t2) = comp_to_array (comp_to_option comp2) n t1 t2 in
    let rec step elts prec equiv =
      match (montaille comp2 comp n a tr' elts prec equiv) with
	| (false,_,_,eq) -> (elts,eq)
	| (_,a,p,e) -> step a p e
    in
    let init,trad = 
      let t' = Array.make n None in
      let m,_ =
	List.fold_left
	  (fun (m,k) q -> t'.(k)<-(Some q); (EMap.add q k m,k+1))
	  (EMap.empty,0)
	  e
      in
      (([],t'),fun q -> EMap.find q m)
    in
    let (elts,eq) = step [init] [init] []
    in
    let elts',eq' =
      (List.sort (fun (u,_) (v,_) -> compmots u v)
	 (List.map (fun (m,s) -> (List.rev m,s)) elts),
       List.map (fun (m,m') -> (List.rev m,List.rev m')) eq)
    in
    let p,l =
      let aux (acc1,acc2) (x,t) = 
	match t.(trad q0) with
	  | None -> (acc1,acc2)
	  | Some q' ->
	    if mem comp2 q' fn
	    then (inser compmots x acc1,inser compmots x acc2)
	    else (acc1,inser compmots x acc2)
      in
      List.fold_left aux ([],[]) elts'
    in
    ((match print with
      | Some (f,g) -> printmon f (option_to_string g) (elts',eq',p)
      | None -> ());
     (p,l,(elts',eq')))

  let monsyn3
      ((e,i,fn,a,tr) : (etat,lettre) automate) 
      : (int -> etat) * (lettre,etat option) monoid =
    let _ = 
      match i with 
	| [q] -> q 
	| _ -> failwith "monsyn : automate non déterministe" 
    in
    let n = List.length e in
    let tr' = 
      let f = makefuncnd(module Etat)(module Alph) tr in
      (fun q a ->
	match f q a with
	  | [] -> None
	  | [q'] -> Some q'
	  | _ -> failwith "caca")
    in
    let comp (_,t1) (_,t2) = comp_to_array (comp_to_option comp2) n t1 t2 in
    let rec step elts prec equiv =
      match (montaille comp2 comp n a tr' elts prec equiv) with
	| (false,_,_,eq) -> (elts,eq)
	| (_,a,p,e) -> step a p e
    in
    let init,trad = 
      let t' = Array.make n None in
      let m,_ =
	List.fold_left
	  (fun (m,k) q -> t'.(k)<-(Some q); (EMap.add q k m,k+1))
	  (EMap.empty,0)
	  e
      in
      (([],t'),fun q -> EMap.find q m)
    in
    let (elts,eq) = step [init] [init] []
    in
    let elts',eq' =
      (List.sort (fun (u,_) (v,_) -> compmots u v)
	 (List.map (fun (m,s) -> (List.rev m,s)) elts),
       List.map (fun (m,m') -> (List.rev m,List.rev m')) eq)
    in
    ((fun i -> match (snd init).(i) with None -> failwith "impossible" | Some q -> q),
     (elts',eq'))

  let clbeta (elts,rules : (lettre,'a) monoid) : mot -> mot =
    let is_in = mem2 compmots in
    let rec clbeta m =
      if (is_in m elts)
      then m
      else (clbeta (beta rules m))
    in
    clbeta

  let prod (m : (lettre,'a) monoid) u v = clbeta m (u@v)

  let prodl (m : (lettre,'a) monoid) u (g : mot list) = 
    let ( ** ) = prod m in
    let rec aux acc = function
      | [] -> acc
      | v::q -> aux ((u**v)|>acc) q
    in
    aux [] g
      
  let prodr (m : (lettre,'a) monoid) (g : mot list) u =
    let ( ** ) = prod m in
    let rec aux acc = function
      | [] -> acc
      | v::q -> aux ((v**u)|>acc) q
    in
    aux [] g

  let prodlst (m : (lettre,'a) monoid) (g : mot list) =
    let ( |* ) = prodr m in
    let rec aux acc = function
      | [] -> acc
      | v::q -> aux (conc compmots [g|*v;acc]) q
    in
    aux []

  let star m g =
    let ( *| ) = prodl m in
    let rec aux g =
      match 
	(List.filter (fun u -> not (is_in u g)) 
	   (bind compmots (fun u -> u*|g) g)) 
      with
	| [] -> g
	| g' -> aux (List.fold_right (|>) g g')
    in
    aux ([]|>g)

  let func_g m (prime : lettre -> lettre) : (mot list -> lettre -> mot list) =
    let ( *| ) = prodl m
    and ( |* ) = prodr m
    in
    (fun g a -> (star m ([prime a] *| (g |* [a]))))

  let auto_g m p a (prime : lettre -> lettre) : (mot list,lettre) automate =
    let final g = inter compmots g p in
    build (module ListB(ListB(Alph))) [[]] final a (func_g m prime)

  let auto_f m p l a (prime : lettre -> lettre) 
      : (mot list * mot list,lettre) automate =
    let ( |* ) = prodr m
    and ( ** ) = prodlst m
    in
    let (eg,_,_,_,tg) = auto_g m [] a prime in
    let fg = 
      let f = makefuncnd(module ListB(ListB(Alph)))(module Alph) tg in
      fun g a ->
	match f g a with
	  | [g'] -> g'
	  | _ -> failwith "auto_f : unknown G-transition"
    in
    let trans (f,g) a = 
      let g' = fg g a in 
      ((f|*[a])**g',g') 
    in
    let final (f,g) = inter compmots f p in
    build 
      (module PairB (ListB(ListB(Alph)))(ListB(ListB(Alph)))) 
      ([[]],[[]]) 
      final 
      a 
      trans

  let cayleyr m p alph =
    (*let p' = List.sort compmots p in*)
    build
      (module Mot)
      []
      (fun w -> mem compmots w p)
      alph
      (fun w a -> prod m w [a])

  let auto_f_opt m p l a (prime : lettre -> lettre) 
      : (mot list * mot list,lettre) automate =
    let ( |* ) = prodr m
    and ( ** ) = prodlst m
    in
    let (eg,_,_,_,tg) = auto_g m [] a prime in
    let restr =
      let (coaccess,_,_,_,_) = coaccess(module Mot)(module Alph)(cayleyr m p a) in
      List.filter (fun w -> List.mem w coaccess)
    in
    let fg = 
      let f = makefuncnd(module ListB(ListB(Alph)))(module Alph) tg in
      fun g a ->
	match f g a with
	  | [g'] -> g'
	  | _ -> failwith "auto_f : unknown G-transition"
    in
    let trans (f,g) a = 
      let g' = fg g a in 
      (restr (f|*[a])**g',g') 
    in
    let final (f,g) = inter compmots f p in
    build 
      (module PairB (ListB(ListB(Alph)))(ListB(ListB(Alph)))) 
      ([[]],[[]]) 
      final 
      a 
      trans

  let auto_f_opt2 m p l a (prime : lettre -> lettre) 
      : (mot list * mot list,lettre) automate =
    let ( |* ) = prodr m
    and ( ** ) = prodlst m
    in
    let (eg,_,_,_,tg) = auto_g m [] a prime in
    let restr =
      let (coaccess,_,_,_,_) = coaccess(module Mot)(module Alph)(cayleyr m p a) in
      let test = 
	(*let l = List.sort compmots coaccess in*)
	fun w -> mem compmots w coaccess
      in
      List.filter test
    in
    let fg = 
      let f = makefuncnd(module ListB(ListB(Alph)))(module Alph) tg in
      fun g a ->
	match f g a with
	  | [g'] -> g'
	  | _ -> failwith "auto_f : unknown G-transition"
    in
    let trans (f,g) a = 
      match f with
	| [] -> ([],[])
	| _ ->
	  begin
	    let g' = fg g a in 
	    match (restr (f|*[a])**g') with 
	      | [] -> ([],[])
	      | f' -> (f',g')
	  end 
    in
    let final (f,g) = inter compmots f p in
    build 
      (module PairB (ListB(ListB(Alph)))(ListB(ListB(Alph)))) 
      ([[]],[[]]) 
      final 
      a 
      trans

  let auto_f_opt3 m p l a (prime : lettre -> lettre) 
      : (mot list * mot list,lettre) automate =
    let ( |* ) = prodr m
    and ( ** ) = prodlst m
    in
    let (eg,_,_,_,tg) = auto_g m [] a prime in
    let restr =
      let test = 
	fun w -> mem compmots w l
      in
      List.filter test
    in
    let fg = 
      let f = makefuncnd(module ListB(ListB(Alph)))(module Alph) tg in
      fun g a ->
	match f g a with
	  | [g'] -> g'
	  | _ -> failwith "auto_f : unknown G-transition"
    in
    let trans (f,g) a = 
      match f with
	| [] -> ([],[])
	| _ ->
	  begin
	    let g' = fg g a in 
	    match (restr (f|*[a])**g') with 
	      | [] -> ([],[])
	      | f' -> (f',g')
	  end 
    in
    let final (f,g) = inter compmots f p in
    build 
      (module PairB (ListB(ListB(Alph)))(ListB(ListB(Alph)))) 
      ([[]],[[]]) 
      final 
      a 
      trans

end
