open Tools
open Automate

type 'a regexp =
| Vide
| Eps
| L of 'a
| Conc of 'a regexp * 'a regexp
| Union of 'a regexp * 'a regexp
| Etoile of 'a regexp

let rec epsilon = function
  | Eps | Etoile _ -> true
  | Vide | L _ -> false
  | Conc (e,f) -> 
    if epsilon e
    then epsilon f
    else false
  | Union (e,f) -> 
    if epsilon e
    then true
    else epsilon f

let rec premier c = function
  | L x -> [x]
  | Eps | Vide -> []
  | Union (e,f) -> conc c [premier c e; premier c f]
  | Conc (e,f) -> 
    begin
      match epsilon e with
      | true -> conc c [premier c e; premier c f]
      | false -> premier c e
    end
  | Etoile e -> premier c e

let rec dernier c = function
  | L x -> [x]
  | Eps | Vide -> []
  | Union (e,f) -> conc c [dernier c e; dernier c f]
  | Conc (e,f) -> 
    begin
      match epsilon f with
      | true -> conc c [dernier c e; dernier c f]
      | false -> dernier c f
    end
  | Etoile e -> dernier c e

let rec suite c a = function
  | L _ | Eps | Vide -> []
  | Union (e,f) -> conc c [suite c a e; suite c a f]
  | Conc (e,f) -> 
    begin
      match (mem c a (dernier c e)) with
      | false -> conc c [suite c a e; suite c a f]
      | true -> conc c [suite c a e; suite c a f; premier c f]
    end
  | Etoile e -> 
      match (mem c a (dernier c e)) with
      | false -> suite c a e
      | true -> conc c [suite c a e;premier c e]

let rec map f = function
  | Vide -> Vide
  | Eps -> Eps
  | L x -> L (f x)
  | Union (e,e') -> Union (map f e,map f e')
  | Conc (e,e') -> Conc (map f e,map f e')
  | Etoile e -> Etoile (map f e) 

let rec fold f i = function
  | Vide -> (Vide,i)
  | Eps -> (Eps,i)
  | L x -> 
    let (a,j) = f i x in 
    (L a,j)
  | Union (e1,e2) -> 
    let (e1',j) = fold f i e1 in 
    let (e2',k) = fold f j e2 in
    (Union (e1',e2'),k)
  | Conc (e1,e2) -> 
    let (e1',j) = fold f i e1 in 
    let (e2',k) = fold f j e2 in
    (Conc (e1',e2'),k)
  | Etoile e -> 
    let (e',j) = fold f i e in 
    (Etoile e',j)

let auto_r 
    (c : 'a -> 'a -> int)
    (a : 'a list)
    (r : 'a regexp)
    : (('a * int) option,'a) automate =
  let etats,r' =
    let index =
      (fun (i,etats) a -> 
	let q = Some(a,i)
	in
	q,(i+1,q::etats))
    in
    let (r',(_,etats)) = fold index (0,[]) r in
    (None::etats,Conc (r',L None))
  in
  let c' = comp_to_option (comp_to_pair c compare) in
  let init = premier c' r'
  and final = [None]
  in
  let trans =
    List.fold_left
      (fun l -> function
      | None -> l
      | Some (a,_) as q ->
	(List.map (fun q' -> (q,a,q')) (suite c' q r'))@l)
      []
      etats
  in
  (etats,init,final,a,trans)

(*t rec automate  
    (c : 'a -> 'a -> int)
    (a : 'a list)
    (r : 'a regexp)
    : (int,'a) automate =*)
 
type 'a revexp =
| RVide
| REps
| RL of 'a
| RConc of 'a revexp * 'a revexp
| RUnion of 'a revexp * 'a revexp
| REtoile of 'a revexp
| Rev of 'a revexp

let rec push_rev prime = function
  | RVide -> Vide
  | REps -> Eps
  | RL a -> L (prime a)
  | RConc (e,f) -> Conc (push_rev prime f,push_rev prime e)
  | RUnion (e,f) -> Union (push_rev prime e,push_rev prime f)
  | REtoile e -> Etoile (push_rev prime e)
  | Rev e -> rev_to_reg prime e
and rev_to_reg prime = function
  | RVide -> Vide
  | REps -> Eps
  | RL a -> L a
  | RConc (e,f) -> Conc (rev_to_reg prime e,rev_to_reg prime f)
  | RUnion (e,f) -> Union (rev_to_reg prime e,rev_to_reg prime f)
  | REtoile e -> Etoile (rev_to_reg prime e)
  | Rev e -> push_rev prime e 
