module Descriptor : UnionFind.Desc with type accumulator = unit = struct
    type descriptor = int
    let default = 0
    type accumulator = unit
    let union d1 d2 acc = ((*print_string "paire ajoutée\n";*)(d1+d2,()))
  end

type ('a,'b) case = A of 'a | B of 'b

let comp_to_list c l1 l2 =
  let rec f = function
    | a::q,b::q' -> 
      (match c a b with
	| 0 -> f (q,q')
	| n -> n)
    | _ -> 0
  in
  match (List.length l1 - List.length l2) with
    | 0 -> f (l1,l2)
    | n -> n

let comp_to_array c n t1 t2 =
  let rec aux = function
    | -1 -> 0
    | k -> 
      (match c t1.(k) t2.(k) with
	| 0 -> aux (k-1)
	| n -> n)
  in
  (aux (n-1))

let comp_to_pair c1 c2 (a,b) (c,d) =
  match c1 a c with
    | 0 -> c2 b d
    | k -> k

let comp_to_pair2 c1 c2 (a,b) (c,d) =
  match c2 b d with
    | 0 -> c1 a c
    | k -> k

let comp_to_tripl c1 c2 c3 (a,b,e) (c,d,f) =
  match c1 a c with
    | 0 -> 
      begin
	match c2 b d with
	  | 0 -> c3 e f
	  | k -> k
      end
    | k -> k

let comp_to_option c a b =
  match (a,b) with
    | None,None -> 0
    | None,_ -> -1
    | _,None -> 1
    | Some a',Some b' -> c a' b'

module type Base = sig
  type t 
  val compare : t -> t -> int
end

module CaseB (B1 : Base) (B2 : Base) : Base
  with type t = (B1.t,B2.t) case = 
struct
  type t = (B1.t,B2.t) case
  let compare a b = 
    match (a,b) with
      | (A a,A a') -> (B1.compare a a')
      | (B b,B b') -> (B2.compare b b')
      | (A _,_) -> -1
      | _ -> 1
end

module ListB (B : Base) : Base
  with type t = B.t list =
struct
  type t = B.t list
  let compare = comp_to_list B.compare
end

module ArrayB (B : Base) : Base
  with type t = B.t array =
struct
  type t = B.t array
  let compare t1 t2 = 
    let n = Array.length t1 in
    match n - Array.length t2 with
      | 0 -> comp_to_array B.compare n t1 t2
      | x -> x
end

module PairB (B1 : Base) (B2 : Base) : Base
  with type t = (B1.t * B2.t) = 
struct
  type t = (B1.t * B2.t)
  let compare = (comp_to_pair B1.compare B2.compare)
end

module Pair2B (B1 : Base) (B2 : Base) : Base
  with type t = (B1.t * B2.t) = 
struct
  type t = (B1.t * B2.t)
  let compare = (comp_to_pair2 B1.compare B2.compare)
end

module OptionB (B : Base) : Base
  with type t = B.t option =
struct
  type t = B.t option
  let compare = comp_to_option B.compare
end

module ToItem (B: Base) = struct
  type t = B.t
  let equal a b = (B.compare a b = 0)
  module Map = Map.Make(B)
end

module MakeUF (B1 : Base) (B2 : Base) : UnionFind.S 
  with 
    type accumulator = unit and type item = (B1.t,B2.t) case =
      UnionFind.Make(ToItem(CaseB(B1)(B2)))(Descriptor)

type ('a,'b) bij = ('a -> 'b) * ('b -> 'a)

let rec mem c a = function
  | [] -> false
  | b::q -> 
    begin
      match c a b with 
	| n when n<0 -> false 
	| 0 -> true 
	| _ -> (mem c a q)
    end

let rec mem2 c a = function
  | [] -> false
  | (b,_)::q -> 
    begin
      match c a b with 
	| n when n<0 -> false 
	| 0 -> true 
	| _ -> (mem2 c a q)
    end

let rec inser c a = function
  | [] -> [a]
  | b::q ->
    begin
      match c a b with 
	| 0 -> b::q
	| n when n < 0 -> a::b::q
	| _ -> b::(inser c a q)
    end

let rec inserb c a = function
  | [] -> Some [a]
  | b::q ->
    begin
      match c a b with 
	| 0 -> None
	| n when n < 0 -> Some (a::b::q)
	| _ -> 
	  begin
	    match (inserb c a q) with
	      | None -> None
	      | Some l -> Some (b::l)
	  end
    end

let inter c l1 l2 =
  let rec aux = function
    | [],_ | _,[] -> false
    | a::q,b::q' ->
      begin 
	match c a b with
	  | 0 -> true
	  | n when n<0 -> aux (q,b::q')
	  | _ -> aux (a::q,q')
      end
  in
  aux (l1,l2)

let makebij (e : 'a list) : ('a,int) bij =
  let n = 
    match List.length e with
      | 0 -> failwith "makemap : empty set"
      | n -> n
  in
  let h = Hashtbl.create n 
  and t = Array.make n (List.hd e)
  and i = ref 0 in
  (List.iter (fun a -> Hashtbl.add h a (!i);t.(!i)<- a;i:=(!i +1)) e;
  (fun a -> Hashtbl.find h a),fun i -> t.(i))

let rec conc c = function
  | [] -> []
  | a::q -> List.fold_right (inser c) (List.fold_right (inser c) a []) (conc c q)

let (@@) l1 l2 =
  let rec aux acc = function
    | [] -> acc
    | a::q -> aux (a::acc) q
  in
  aux l2 l1

let bind c f l =  conc c (List.map f l)

let rec assoc c x = function
  | (y,t)::q ->
    begin 
      match c x y with
	| 0 -> t
	| 1 -> assoc c x q
	| _ -> failwith "assoc : not found1"
    end
  | _ -> failwith "assoc : not found2"

let rec assoc2 c x = function
  | (y,t)::q ->
    begin 
      match c x y with
	| 0 -> t::(assoc2 c x q)
	| 1 -> assoc2 c x q
	| _ -> []
    end
  | _ -> []

let makefuncnd
    (type a) 
    (type b)
    (module B1 : Base with type t = a)
    (module B2 : Base with type t = b)
    (list : (a * b * a) list)
    : a -> b -> a list =
  let module Map = Map.Make(PairB(B1)(B2)) in
  let m =
    List.fold_left 
      (fun m (q,a,q') -> (fun x y m -> 
	try Map.add x (y::(Map.find x m)) m 
	with Not_found -> Map.add x [y] m)
	(q,a) q' m)
      Map.empty
      list
  in
  (fun q a ->
    try (Map.find (q,a) m)
    with Not_found -> [])

let makefuncd
    (type a) 
    (type b)
    (module B1 : Base with type t = a)
    (module B2 : Base with type t = b)
    (list : (a * b * a) list)
    : a -> b -> a option =
  let module Map = Map.Make(PairB(B1)(B2)) in
  let m =
    List.fold_left 
      (fun m (q,a,q') -> (fun x y m -> 
	try 
	  let _ = (Map.find x m) in
	  failwith "makefuncd : Automate non déterministe." 
	with Not_found -> Map.add x (Some y) m)
	(q,a) q' m)
      Map.empty
      list
  in
  (fun q a ->
    try (Map.find (q,a) m)
    with Not_found -> None)

let mappairs c f base =
  let rec aux acc1 acc2 = function
    | [] -> (acc1,acc2)
    | a::q -> 
      let b1,b2 = f a in
      aux (inser c b1 acc1) (b2::acc2) q
  in
  aux [] base

let (==) l1 l2 =
  let l1' = List.filter (fun x -> List.mem x l2) l1 
  and l2' = List.filter (fun x -> List.mem x l1) l2 in
  (l1=l1')&(l2=l2')

let (^^) l1 l2 =
  List.fold_left 
    (fun l x -> if List.mem x l then l else x::l) 
    [] 
    (l1@l2)

let rec remove c x = function
  | [] -> []
  | y::q ->
    begin
      match c x y with
	| 0 -> q
	| k when k<0 -> y::q
	| _ -> y::(remove c x q)
    end

let chrono f x =
  let t = Unix.gettimeofday () in
  let r = (f x) in
  Printf.printf "Temps écoulé : %fs\n" (Unix.gettimeofday () -. t);
  r

let fold_lefti f acc tab =
  let (_,acc') = 
    Array.fold_left
      (fun (i,acc) x -> (i+1,f i acc x))
      (0,acc)
      tab
  in
  acc'
